

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>

#include "pac7311-grab-cmd-handlers.h"
#include "pac7311-grab-jpeg.h"
#include "pac7311-stream.h"


cmd_list_type command_list = {

	//			  no args
	{ "start_stream",	{ CMD_ARG_END },
		cmd_start_stream },
	
	//			  no args
	{ "stop_stream",	{ CMD_ARG_END },
		cmd_stop_stream },
	
	//			  register       value
	{ "set_reg",		{ CMD_ARG_UCHAR, CMD_ARG_UCHAR, CMD_ARG_END },
		cmd_set_reg },
	
	//			  register
	{ "print_reg",		{ CMD_ARG_UCHAR, CMD_ARG_END },
		cmd_print_reg },
	
	//			  index          message          message size
	{ "send_ctrl_msg",	{ CMD_ARG_UCHAR, CMD_ARG_HEX_STR, CMD_ARG_UCHAR, CMD_ARG_END },
		cmd_send_ctrl_msg },
	
	//			  useconds
	{ "usleep",		{ CMD_ARG_ULONG, CMD_ARG_END },
		cmd_usleep },

	//			  no args
	{ "grab_frame",	{ CMD_ARG_END },
		cmd_grab_frame },

	//			  no of tries    delay (usec)
	{ "grab_frame_burst",	{ CMD_ARG_UCHAR, CMD_ARG_ULONG, CMD_ARG_END },
		cmd_grab_frame_burst },

	//			  file name
	{ "header_to_file",	{ CMD_ARG_STR, CMD_ARG_END },
		cmd_header_to_file },
		
	//			  width          height         file name
	{ "jpeg_to_file",	{ CMD_ARG_ULONG, CMD_ARG_ULONG, CMD_ARG_STR, CMD_ARG_END },
		cmd_jpeg_to_file },
	
	//			  file name
	{ "raw_to_file",	{ CMD_ARG_STR, CMD_ARG_END },
		cmd_raw_to_file },
		
	//			  string
	{ "println",		{ CMD_ARG_STR, CMD_ARG_END },
		cmd_println },
		
	{ CMD_END } };



unsigned int pac7311_grab_frame_size = 0;
unsigned char pac7311_grab_frame_data[PAC7311_GRAB_MAX_FRAME_SIZE];


unsigned char pac7311_grab_header_marker[] = { 0x1a, 0xff, 0xff, 0x00, 0xff, 0x96, 0x62 };
unsigned char pac7311_grab_header_marker_len = 7;



int __find_frame_header_end()
{
	int search_len = pac7311_grab_frame_size - pac7311_grab_header_marker_len + 1;
	int i;
	
	if (search_len <= 0)
	{
		return -1;
	}
	
	for (i = 0; i < search_len; ++i)
	{
		if (!memcmp(pac7311_grab_frame_data + i, pac7311_grab_header_marker,
		            pac7311_grab_header_marker_len))
		{
			break;
		}
	}
	
	if (i < search_len)
	{
		return (i + pac7311_grab_header_marker_len);
	}
	else
	{
		return -1;
	}
}



int cmd_start_stream(int pac7311_fd, arg_value_type arg_values[])
{
	PFUNC_CALLED;


	return pac7311_stream_stream_start(pac7311_fd);
}


int cmd_stop_stream(int pac7311_fd, arg_value_type arg_values[])
{
	PFUNC_CALLED;


	return pac7311_stream_stream_stop(pac7311_fd);
}


int cmd_set_reg(int pac7311_fd, arg_value_type arg_values[])
{
	unsigned char reg = arg_values[0].uchar;
	unsigned char value = arg_values[1].uchar;

	PFUNC_CALLED;


	return pac7311_stream_reg_set(pac7311_fd, reg, value);
}


int cmd_print_reg(int pac7311_fd, arg_value_type arg_values[])
{
	int retval;

	unsigned char reg = arg_values[0].uchar;
	unsigned char value;

	PFUNC_CALLED;


	retval = pac7311_stream_reg_get(pac7311_fd, reg, &value);

	if (!retval)
	{
		printf("Register 0x%02x has value 0x%02x.\n", reg, value);
	}

	return retval;
}


int cmd_send_ctrl_msg(int pac7311_fd, arg_value_type arg_values[])
{
	unsigned char index = arg_values[0].uchar;
	unsigned char *data = arg_values[1].hex_str;
	unsigned char data_len = arg_values[2].uchar;

	PFUNC_CALLED;


	return pac7311_stream_control_msg_send(pac7311_fd, index, data, data_len);
}


int cmd_usleep(int pac7311_fd, arg_value_type arg_values[])
{
	unsigned long usecs = arg_values[0].ulong;
	
	PFUNC_CALLED;


	usleep(usecs);

	return 0;
}


int cmd_grab_frame(int pac7311_fd, arg_value_type arg_values[])
{
	int bytes_read;
	
	PFUNC_CALLED;


	pac7311_grab_frame_size = 0;

	bytes_read = read(pac7311_fd, pac7311_grab_frame_data, PAC7311_GRAB_MAX_FRAME_SIZE);
	
	if (bytes_read <= 0)
	{
		return 1;
	}
	
	pac7311_grab_frame_size = bytes_read;
	
	return 0;
}


int cmd_grab_frame_burst(int pac7311_fd, arg_value_type arg_values[])
{
	unsigned char tries = arg_values[0].uchar;
	unsigned long delay = arg_values[1].ulong;

	int bytes_read = 0;
	
	PFUNC_CALLED;


	pac7311_grab_frame_size = 0;

	while ((bytes_read <= 0) &&
	       (tries--))
	{
		bytes_read = read(pac7311_fd, pac7311_grab_frame_data, PAC7311_GRAB_MAX_FRAME_SIZE);
		
		if (bytes_read <= 0)
			usleep(delay);
	}
	
	if (bytes_read <= 0)
	{
		return 1;
	}
	
	pac7311_grab_frame_size = bytes_read;
	
	return 0;
}


int cmd_jpeg_to_file(int pac7311_fd, arg_value_type arg_values[])
{
	unsigned long image_width = arg_values[0].ulong;
	unsigned long image_height = arg_values[1].ulong;
	char *filename = arg_values[2].str;
	
	FILE *file;
	int header_size, jpeg_header_size;
	int retval = 0;
	unsigned char *jpeg_header_data;
	unsigned char jpeg_mcu_marker = 0x00;
	
	PFUNC_CALLED;


	if (pac7311_grab_frame_size == 0)
	{
		retval = 1;
		goto exit0;
	}
	
	if ((header_size = __find_frame_header_end()) < 0)
	{
		retval = 2;
		goto exit0;
	}

	jpeg_header_data = pac7311_jpeg_get_header(image_width, image_height, &jpeg_header_size);

	if (pac7311_grab_frame_size > header_size)
	{
		jpeg_mcu_marker = pac7311_grab_frame_data[header_size];
	}

	if (!(file = fopen(filename, "w")))
	{
		retval = 3;
		goto exit0;
	}

	
	if (fwrite(jpeg_header_data, jpeg_header_size, 1, file) != 1)
	{
		retval = 4;
		goto exit1;
	}

	if (pac7311_jpeg_write_jpeg(
		pac7311_grab_frame_data + header_size, pac7311_grab_frame_size - header_size,
		image_width, image_height,
		jpeg_mcu_marker, file))
	{
		retval = 5;
		goto exit1;
	}
	
	if (fwrite("\xff\xd9", 2, 1, file) != 1)
	{
		retval = 6;
		goto exit1;
	}

exit1:
	fclose(file);

exit0:	
	return retval;
}


int cmd_header_to_file(int pac7311_fd, arg_value_type arg_values[])
{
	char *filename = arg_values[0].str;
	
	FILE *file;
	int header_size;
	int retval = 0;
	
	PFUNC_CALLED;

	if (pac7311_grab_frame_size == 0)
	{
		return 1;
	}
	
	if ((header_size = __find_frame_header_end()) < 0)
	{
		return 2;
	}
	
	if (!(file = fopen(filename, "w")))
	{
		return 3;
	}

	if (fwrite(pac7311_grab_frame_data, header_size, 1, file) != header_size)
	{
		retval = 4;
	}
	
	fclose(file);
	
	return 0;
}


int cmd_raw_to_file(int pac7311_fd, arg_value_type arg_values[])
{
	char *filename = arg_values[0].str;
	
	FILE *file;
	int retval = 0;
	
	PFUNC_CALLED;


	if (pac7311_grab_frame_size == 0)
	{
		retval = 1;
		goto exit0;
	}
	
	if (!(file = fopen(filename, "w")))
	{
		retval = 2;
		goto exit0;
	}
	
	if (fwrite(pac7311_grab_frame_data, pac7311_grab_frame_size, 1, file) != 1)
	{
		retval = 3;
		goto exit1;
	}
	
exit1:
	fclose(file);

exit0:	
	return retval;
}


int cmd_println(int pac7311_fd, arg_value_type arg_values[])
{
	char *string = arg_values[0].str;
	
	PFUNC_CALLED;
	
	
	printf("%s\n", string);
	
	return 0;
}
