#!/usr/bin/perl -w
# Please send any comments, bug reports or new ideas to timo.lindfors@iki.fi
# Please see http://iki.fi/lindi/usb/usbsnoop.txt for usage instructions.
# 2006-08-30/lindi: handle URB_FUNCTION_SELECT_INTERFACE, escape \0 in perl
# 2006-07-18/lindi: allow product id to be 0x0000
# 2006-03-09/lindi: don't fill buffer before reads
# 2005-09-18/lindi: parse transfer buffers properly
# 2005-12-24/lindi: 
use strict;

my ($line,$urb, $last_time);

my %endpointtype;

$last_time = 0;

# Note that if you enable support for isochronous transfers you also
# need the following two files:
# http://iki.fi/lindi/usb/libusb_augment.c
# http://iki.fi/lindi/usb/libusb_augment.h
my $useisoch = 1;

my $usbsnoopbugseen = 0;

sub get_transferbuffer_contents {
    my (@lines) = @_;
    my $buffer = "";
    foreach my $line (@lines) {
	chomp($line);
	if ($line =~ /^    0[0-9a-f]{7}: ([^ ].+)/) {
	    my $bytes = $1;
#	    $bytes =~ s/ /\\x/g;
	    $bytes =~ s/ //g;
#	    $bytes =~ s/^/\\x/;
	    $buffer .= $bytes;
	}
    }
    return $buffer;
}

sub process_urb {
    my $text = shift;
    if (!defined($text)) {
	return;
    }
    my @lines = split(/\n/, $text);
    if ($text =~ m/UsbSnoop - incorrect UrbHeader.Length/) {
	print("OTHER: usbsnoop bug detected\n");
	$usbsnoopbugseen = 1;
    }
    if ($text =~ m/coming back/) {
	if ($text =~ m/URB_FUNCTION_SELECT_CONFIGURATION/) {
	    my $endpoint;
	    foreach my $line (@lines) {
		if ($line =~ /EndpointAddress *= 0x([a-f0-9]+)/) {
		    $endpoint = $1;
		}
		if ($line =~ /PipeType *= 0x([a-f0-9]+) /) {
		    my $pipetype = $1;
		    $endpointtype{$endpoint} = $pipetype;
		}
	    }
	} elsif (($usbsnoopbugseen == 1) and ($text =~ m/URB_FUNCTION_SELECT_INTERFACE/)) {
	    # UsbSnoop - incorrect UrbHeader.Length=0, should be at least 16
	    # and usbsnoop does not show what we sent! So, we need to parse
	    # alternate interface from the reply.
	    if ($text =~ m/Interface: AlternateSetting *= ([0-9]+)/) {
		my $AlternateSetting = $1;
		chomp($AlternateSetting);
		if ($text =~ /URB (\d+) /) {
		    my $urbnumber = $1;
		    chomp($urbnumber);
#		    print "ret = usb_set_altinterface(devh, $AlternateSetting);\n";
#		    print "printf(\"$urbnumber set alternate setting returned %d\\n\", ret);\n";
		    print "OTHER: usb_set_altinterface\n";
		}
	    }
	    $usbsnoopbugseen = 0;
	}
	return;
    }
#    print "text=\"$text\"\n";
    my ($time, $TransferBufferLength, $DescriptorType, $Index, $bConfigurationValue, $Value, $Request, $endpoint, $urbnumber, $AlternateSetting, $FeatureSelector, $IsoPacket1Offset, $IsoPacketLastIndex);
    if ($text =~ m/(\d+) ms/) {
	$time = $1;
    }

    if ($last_time == 0) {
	$last_time = $time;
    } else {
	my $timediff = $time - $last_time;
	if ($timediff > 2500) {
	    $timediff = 2500; # FIXME
	}
	if ($timediff > 0) {
#	    print "usleep($timediff*1000);\n";
	}
	$last_time = $time;
    }
    if ($text =~ m/TransferBufferLength = 0([^ ]+)/) {
	$TransferBufferLength = $1;
	chomp($TransferBufferLength);
    }
    if ($text =~ m/DescriptorType *= 0([^ ]+)/) {
	$DescriptorType = $1;
	chomp($DescriptorType);
    }
    if ($text =~ m/Index *= 0([a-f0-9]+)/) {
	$Index = $1;
	chomp($Index);
    }
    if ($text =~ m/bConfigurationValue *= 0x0([^ ]+)/) {
	$bConfigurationValue = $1;
	chomp($bConfigurationValue);
    }
    if ($text =~ m/Value *= 0([^ ]+)/) {
	$Value = $1;
	chomp($Value);
    }
    if ($text =~ m/Request *= 0([^ ]+)/) {
	$Request = $1;
	chomp($Request);
    }
    if ($text =~ m/endpoint 0x([^ \]]+)/) {
	$endpoint = $1;
	chomp($endpoint);
    }
    if ($text =~ m/URB (\d+) /) {
	$urbnumber = $1;
	chomp($urbnumber);
    }
    if ($text =~ m/AlternateSetting *= ([0-9]+)/) {
	$AlternateSetting = $1;
	chomp($AlternateSetting);
    }
    if ($text =~ m/FeatureSelector *= ([0-9]+)/) {
	$FeatureSelector = $1;
	chomp($FeatureSelector);
    }
    if ($text =~ m/IsoPacket\[1\].Offset *= (\d+)/) {
	$IsoPacket1Offset = $1;
	chomp($IsoPacket1Offset);
    }
    foreach my $line (@lines) {
	if ($line =~ /IsoPacket\[(\d+)\]/) {
	    $IsoPacketLastIndex = $1;
	    chomp($IsoPacketLastIndex);
	}
    }
    if ($text =~ m/URB_FUNCTION_GET_DESCRIPTOR_FROM_DEVICE/) {
#	print "ret = usb_get_descriptor(devh, 0x$DescriptorType, 0x$Index, buf, 0x$TransferBufferLength);\n";
#	print "printf(\"$urbnumber get descriptor returned %d, bytes: \", ret);\n";
#	print "print_bytes(buf, ret);\n";
#	print "printf(\"\\n\");\n";
        print "OTHER: usb_get_descriptor (device)\n";
    } elsif ($text =~ m/URB_FUNCTION_GET_DESCRIPTOR_FROM_INTERFACE/) {
#	printf "ret = usb_get_descriptor(devh, 0x$DescriptorType, 0x$Index, buf, 0x$TransferBufferLength);\n";
#	print "printf(\"$urbnumber get descriptor returned %d, bytes: \", ret);\n";
#	print "print_bytes(buf, ret);\n";
#	print "printf(\"\\n\");\n";
        print "OTHER: usb_get_descriptor (interface)\n";
    } elsif ($text =~ m/URB_FUNCTION_SELECT_CONFIGURATION/) {
	if (defined($bConfigurationValue)) {
#	    print "ret = usb_release_interface(devh, 0);\n";
#	    print "if (ret != 0) printf(\"failed to release interface before set_configuration: %d\\n\", ret);\n";
#	    print "ret = usb_set_configuration(devh, 0x$bConfigurationValue);\n";
#	    print "printf(\"$urbnumber set configuration returned %d\\n\", ret);\n";
#	    print "ret = usb_claim_interface(devh, 0);\n";
#	    print "if (ret != 0) printf(\"claim after set_configuration failed with error %d\\n\", ret);\n";
            print "OTHER: usb_set_configuration\n";
	}
	if (defined($AlternateSetting)) {
#	    print "ret = usb_set_altinterface(devh, $AlternateSetting);\n";
#	    print "printf(\"$urbnumber set alternate setting returned %d\\n\", ret);\n";
            print "OTHER: usb_set_altinterface\n";
	}    
    } elsif ($text =~ m/URB_FUNCTION_CLASS_INTERFACE/) {
	my $requesttype = "USB_TYPE_CLASS + USB_RECIP_INTERFACE";
	if ($text =~ m/USBD_TRANSFER_DIRECTION_IN/) {
	    $requesttype .= " + USB_ENDPOINT_IN";
	}
	my $bytes = get_transferbuffer_contents(@lines);
	if ($bytes ne "") {
#	    print "memcpy(buf, \"$bytes\", 0x$TransferBufferLength);\n";
	}
#	print "ret = usb_control_msg(devh, $requesttype, 0x$Request, 0x$Value, 0x$Index, buf, 0x$TransferBufferLength, 1000);\n";

	if ($text =~ m/USBD_TRANSFER_DIRECTION_IN/) {
#	    print "printf(\"$urbnumber control msg returned %d, bytes: \", ret);\nprint_bytes(buf, ret);\n";
	} else {
#	    print "printf(\"$urbnumber control msg returned %d\", ret);\n";
	}
#	print "printf(\"\\n\");\n";
        print "OTHER: usb_control_msg (USB_TYPE_CLASS + USB_RECIP_INTERFACE)\n";
    } elsif ($text =~ m/URB_FUNCTION_VENDOR_INTERFACE/) {
	my $requesttype = "USB_TYPE_VENDOR + USB_RECIP_INTERFACE";
	if ($text =~ m/USBD_TRANSFER_DIRECTION_IN/) {
	    $requesttype .= " + USB_ENDPOINT_IN";
	}
	my $bytes = get_transferbuffer_contents(@lines);
	if ($bytes ne "") {
#	    print "memcpy(buf, \"$bytes\", 0x$TransferBufferLength);\n";
	}
#	print "ret = usb_control_msg(devh, $requesttype, 0x$Request, 0x$Value, 0x$Index, buf, 0x$TransferBufferLength, 1000);\n";
#	print "printf(\"$urbnumber control msg returned %d, bytes: \", ret);\n";
#	print "print_bytes(buf, ret);\n";
#	print "printf(\"\\n\");\n";

        print "OTHER: URB_FUNCTION_VENDOR_INTERFACE\n";

    } elsif ($text =~ m/URB_FUNCTION_BULK_OR_INTERRUPT_TRANSFER/) {
	my $method;
	if ( ( hex $endpoint & 128 ) == 128 ) {
	    # if ($text =~ /USBD_TRANSFER_DIRECTION_IN/) {
	    $method = "read";
	} else {
	    $method = "write";
	}
	my $type = $endpointtype{$endpoint};
	if (!defined($type)) {
	    die "can't find \"$endpoint\"\n";
	}
	my $mode;
	if ($type eq "00000002") {
	    $mode = "bulk";
	} elsif ($type eq "00000003") {
	    $mode = "interrupt";
	} else {
	    die "unrecognised endpointtype \"$type\" for endpoint \"$endpoint\"";
	}
	my $bytes = get_transferbuffer_contents(@lines);
	if ($bytes ne "" and $method eq "write") {
#	    print "memcpy(buf, \"$bytes\", 0x$TransferBufferLength);\n";
	}
	my $timeout = 1000;
	if ($mode eq "bulk" and hex($TransferBufferLength) > 50) {
	    $timeout = 1000 + hex($TransferBufferLength) * 0.06;
	    $timeout = sprintf("%.0d", $timeout); # round to integer
	}
#	print "ret = usb_".$mode."_$method(devh, 0x$endpoint, buf, 0x$TransferBufferLength, $timeout);\n";
#	print "printf(\"$urbnumber $mode $method returned %d, bytes: \", ret);\n";
# sounds weird but write requests can actually also return data
#	if ($method eq "read") {
#	    print "print_bytes(buf, ret);\n";
#	}
#	print "printf(\"\\n\");\n";

        print "OTHER: usb_".$mode."_$method\n";

    } elsif ($text =~ m/URB_FUNCTION_SET_FEATURE_TO_DEVICE/) {
	my $requesttype = "USB_TYPE_STANDARD + USB_RECIP_DEVICE";
#	print "ret = usb_control_msg(devh, $requesttype, USB_REQ_SET_FEATURE, $FeatureSelector, 0, buf, 0, 1000);\n";
#	print "printf(\"$urbnumber set feature request returned %d\\n\", ret);\n";
        print "OTHER: URB_FUNCTION_SET_FEATURE_TO_DEVICE\n";
    } elsif ($text =~ m/URB_FUNCTION_VENDOR_DEVICE/) {
	my $requesttype = "USB_TYPE_VENDOR + USB_RECIP_DEVICE";
	if ($text =~ m/USBD_TRANSFER_DIRECTION_IN/) {
	    $requesttype .= " + USB_ENDPOINT_IN";
	}
	my $bytes = get_transferbuffer_contents(@lines);
	if ($bytes ne "") {
#	    print "memcpy(buf, \"$bytes\", 0x$TransferBufferLength);\n";
	}
#	print "ret = usb_control_msg(devh, $requesttype, 0x$Request, 0x$Value, 0x$Index, buf, 0x$TransferBufferLength, 1000);\n";
#	print "printf(\"$urbnumber control msg returned %d, bytes: \", ret);\n";
#	print "print_bytes(buf, ret);\n";
#	print "printf(\"\\n\");\n";
      	if ($text =~ m/USBD_TRANSFER_DIRECTION_IN/) {
          if ($TransferBufferLength == 1)
          {
            print "print_reg 0x$Index\n";
          }
          else
          {      	
	    print "OTHER: usb_control_msg (USB_TYPE_VENDOR + USB_RECIP_DEVICE + USB_ENDPOINT_IN)\n";
	  }
	}
	else
	{
          if ($TransferBufferLength == 1)
          {
            print "set_reg 0x$Index 0x$Value\n";
          }
          else
          {
            print "send_ctrl_msg 0x$Index $bytes 0x$TransferBufferLength\n";
          }
        }

    } elsif ($text =~ m/URB_FUNCTION_GET_CURRENT_FRAME_NUMBER/) {
	my $requesttype = "USB_TYPE_VENDOR + USB_RECIP_DEVICE";
	# TODO
        print "OTHER: URB_FUNCTION_GET_CURRENT_FRAME_NUMBER\n";
    } elsif ($text =~ m/URB_FUNCTION_ISOCH_TRANSFER/) {
	if ($useisoch == 0) {
	    die "please set useisoch = 1 in the source code to enable experimental isochronous support";
	}
	if (!defined($IsoPacket1Offset)) {
	    die "can't find offset of first iso packet\n";
	}
	if (!defined($IsoPacketLastIndex)) {
	    die "can't find index of last iso packet\n";
	}
	my $packetsize = $IsoPacket1Offset;
	my $packetcount = $IsoPacketLastIndex + 1;
	if ($packetcount * $packetsize > 32768) {
	    die "packetcount ($packetcount) or packetsize ($packetsize) is way too large!\n";
	}
#	print "ret = usb_isochronous_setup(&isourb, 0x$endpoint, $packetsize, isobuf, $packetcount * $packetsize);\n";
#	print "printf(\"$urbnumber isochronous setup returned %d\\n\", ret);\n";
#	print "ret = usb_isochronous_submit(devh, isourb, &isotv);\n";
#	print "printf(\"$urbnumber isochronous submit returned %d\\n\", ret);\n";
#	print "ret = usb_isochronous_reap(devh, isourb, &isotv, 1000);\n";
#	print "printf(\"$urbnumber isochronous reap returned %d, bytes: \", ret);\n";
#	print "print_bytes_isoch(isourb, ret, $packetsize);\n";
#	print "printf(\"\\n\");\n";
	# TODO
	print "OTHER: isochronous transfer\n";
    } elsif ($text =~ m/URB_FUNCTION_ABORT_PIPE/) {
	# TODO: implement
#	print "printf(\"$urbnumber abort pipe not implemented yet :(\\n\");\n";
        print "OTHER: abort pipe\n";
    } elsif ($text =~ m/URB_FUNCTION_RESET_PIPE/) {
	# TODO: implement
#	print "printf(\"$urbnumber reset pipe not implemented yet :(\\n\");\n";
        print "OTHER: reset pipe\n";
    } elsif ($text =~ m/URB_FUNCTION_SELECT_INTERFACE/) {
        if (!defined($AlternateSetting)) {
            die "can't find alternatesetting\n";
        }
#        print "ret = usb_set_altinterface(devh, $AlternateSetting);\n";
#        print "printf(\"$urbnumber set alternate setting returned %d\\n\", ret);\n";
        print "OTHER: usb_set_altinterface\n";
    } elsif ($text =~ m/incorrect UrbHeader.Length=0,/) {
	# ignore
    } else {
	die "unrecognized URB type, text = \"$text\"";
    }
}


while (defined($line = <>)) {
    if ($line =~ m/ URB (\d+) (going down|coming back)/) {
        &process_urb($urb);
        $urb = $line;
    } elsif (defined($urb)) {
        $urb .= $line;
    }
}


&process_urb($urb);

#print <<EOF
#	ret = usb_release_interface(devh, 0);
#	assert(ret == 0);
#	ret = usb_close(devh);
#	assert(ret == 0);
#	return 0;
#}
#EOF
#    ;
